"""Shared test fixtures for the pgns SDK."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

import httpx

from pgns.sdk.async_client import AsyncPigeonsClient
from pgns.sdk.client import PigeonsClient

BASE_URL = "https://api.pgns.io"
API_KEY = "pk_live_test1234567890abcdefghijklmnopqrstuvwxyz1234"


def make_client(handler: Callable[[httpx.Request], httpx.Response]) -> PigeonsClient:
    """Create a sync PigeonsClient backed by a mock transport."""
    transport = httpx.MockTransport(handler)
    http_client = httpx.Client(transport=transport)
    return PigeonsClient(BASE_URL, api_key=API_KEY, http_client=http_client)


def make_async_client(
    handler: Callable[[httpx.Request], httpx.Response],
) -> AsyncPigeonsClient:
    """Create an async PigeonsClient backed by a mock transport."""
    transport = httpx.MockTransport(handler)
    http_client = httpx.AsyncClient(transport=transport)
    return AsyncPigeonsClient(BASE_URL, api_key=API_KEY, http_client=http_client)


SAMPLE_ROOST: dict[str, Any] = {
    "id": "roost_abc123",
    "name": "Test Roost",
    "description": "A test roost",
    "secret": None,
    "is_active": True,
    "created_at": "2024-01-01T00:00:00Z",
    "updated_at": "2024-01-01T00:00:00Z",
}

SAMPLE_PIGEON: dict[str, Any] = {
    "id": "pgn_abc123",
    "roost_id": "roost_abc123",
    "source_ip": "127.0.0.1",
    "request_method": "POST",
    "content_type": "application/json",
    "headers": {"content-type": "application/json"},
    "body_json": {"hello": "world"},
    "body_raw": None,
    "request_query": None,
    "replayed_from": None,
    "delivery_status": "delivered",
    "received_at": "2024-01-01T00:00:00Z",
}

SAMPLE_DESTINATION: dict[str, Any] = {
    "id": "dest_abc123",
    "roost_id": "roost_abc123",
    "name": "My Webhook",
    "destination_type": "url",
    "config": {"url": "https://example.com/webhook"},
    "filter_expression": "",
    "template": "",
    "retry_max": 5,
    "retry_delay_ms": 1000,
    "retry_multiplier": 2.0,
    "is_paused": False,
    "is_verified": True,
    "created_at": "2024-01-01T00:00:00Z",
    "updated_at": "2024-01-01T00:00:00Z",
}

SAMPLE_API_KEY: dict[str, Any] = {
    "id": "key_abc123",
    "key_prefix": "pk_live_test1234",
    "name": "Default",
    "last_used": None,
    "revoked_at": None,
    "created_at": "2024-01-01T00:00:00Z",
}

SAMPLE_USER: dict[str, Any] = {
    "id": "user_abc123",
    "email": "test@example.com",
    "name": "Test User",
    "plan": "free",
    "created_at": "2024-01-01T00:00:00Z",
    "updated_at": "2024-01-01T00:00:00Z",
}

SAMPLE_TEMPLATE: dict[str, Any] = {
    "id": "tmpl_abc123",
    "name": "Test Template",
    "description": "A test template",
    "body": "{{ body }}",
    "created_at": "2024-01-01T00:00:00Z",
    "updated_at": "2024-01-01T00:00:00Z",
}



"""pgns SDK — Python client for the pgns webhook relay API."""

from pgns.sdk._version import __version__
from pgns.sdk.async_client import AsyncPigeonsClient
from pgns.sdk.client import PigeonsClient
from pgns.sdk.errors import PigeonsAuthError, PigeonsError, WebhookVerificationError
from pgns.sdk.events import async_event_stream, event_stream
from pgns.sdk.models import (
    ApiKeyCreatedResponse,
    ApiKeyResponse,
    AuthTokens,
    CreateApiKeyRequest,
    CreateDestination,
    CreateRoost,
    CreateTemplate,
    DeliveryAttempt,
    DeliveryStatus,
    Destination,
    DestinationType,
    PaginatedDeliveryAttempts,
    PaginatedPigeons,
    PauseResponse,
    Pigeon,
    PreviewTemplateRequest,
    PreviewTemplateResponse,
    ReplayResponse,
    Roost,
    SendResponse,
    Template,
    UpdateApiKeyRequest,
    UpdateDestination,
    UpdateProfileRequest,
    UpdateRoost,
    UpdateTemplate,
    User,
)
from pgns.sdk.webhook import Webhook

__all__ = [
    "__version__",
    # Clients
    "PigeonsClient",
    "AsyncPigeonsClient",
    # Errors
    "PigeonsError",
    "PigeonsAuthError",
    "WebhookVerificationError",
    # Webhook
    "Webhook",
    # Events
    "event_stream",
    "async_event_stream",
    # Enums
    "DestinationType",
    "DeliveryStatus",
    # Auth
    "AuthTokens",
    # Domain models
    "User",
    "Roost",
    "Pigeon",
    "Destination",
    "DeliveryAttempt",
    # API Keys
    "ApiKeyResponse",
    "ApiKeyCreatedResponse",
    # Mutations
    "CreateRoost",
    "UpdateRoost",
    "CreateDestination",
    "UpdateDestination",
    "UpdateProfileRequest",
    "CreateApiKeyRequest",
    "UpdateApiKeyRequest",
    # Responses
    "ReplayResponse",
    "PauseResponse",
    "SendResponse",
    # Pagination
    "PaginatedPigeons",
    "PaginatedDeliveryAttempts",
    # Templates
    "Template",
    "CreateTemplate",
    "UpdateTemplate",
    "PreviewTemplateRequest",
    "PreviewTemplateResponse",
]



"""Pydantic v2 models for the pgns API."""

from __future__ import annotations

from enum import StrEnum
from typing import Any

from pydantic import BaseModel

# ---------------------------------------------------------------------------
# Enums
# ---------------------------------------------------------------------------


class DestinationType(StrEnum):
    url = "url"
    slack = "slack"
    discord = "discord"
    email = "email"


class DeliveryStatus(StrEnum):
    pending = "pending"
    delivering = "delivering"
    delivered = "delivered"
    failed = "failed"
    retrying = "retrying"


# ---------------------------------------------------------------------------
# Auth
# ---------------------------------------------------------------------------


class AuthTokens(BaseModel):
    access_token: str
    token_type: str
    expires_in: int


# ---------------------------------------------------------------------------
# Domain models
# ---------------------------------------------------------------------------


class User(BaseModel):
    id: str
    email: str
    name: str
    plan: str
    created_at: str
    updated_at: str


class SourceType(StrEnum):
    github = "github"
    stripe = "stripe"
    shopify = "shopify"
    slack = "slack"
    discord = "discord"
    svix = "svix"
    pigeon = "pigeon"
    linear = "linear"
    sentry = "sentry"


class Roost(BaseModel):
    id: str
    name: str
    description: str
    secret: str | None = None
    source_type: SourceType | None = None
    is_active: bool
    created_at: str
    updated_at: str


class Pigeon(BaseModel):
    id: str
    roost_id: str
    source_ip: str
    request_method: str
    content_type: str
    headers: dict[str, Any]
    body_json: Any | None = None
    body_raw: list[int] | None = None
    request_query: dict[str, Any] | None = None
    replayed_from: str | None = None
    delivery_status: DeliveryStatus
    received_at: str


class Destination(BaseModel):
    id: str
    roost_id: str
    name: str
    destination_type: DestinationType
    config: dict[str, Any]
    filter_expression: str
    template: str
    retry_max: int
    retry_delay_ms: int
    retry_multiplier: float
    is_paused: bool
    is_verified: bool
    created_at: str
    updated_at: str


class DeliveryAttempt(BaseModel):
    id: str
    pigeon_id: str
    destination_id: str
    status: DeliveryStatus
    attempt_number: int
    response_status: int | None = None
    response_body: str | None = None
    response_headers: dict[str, str] | None = None
    error_message: str | None = None
    attempted_at: str
    next_retry_at: str | None = None


# ---------------------------------------------------------------------------
# API Keys
# ---------------------------------------------------------------------------


class ApiKeyResponse(BaseModel):
    id: str
    key_prefix: str
    name: str
    last_used: str | None = None
    revoked_at: str | None = None
    created_at: str


class ApiKeyCreatedResponse(BaseModel):
    id: str
    key: str
    key_prefix: str
    name: str
    created_at: str


# ---------------------------------------------------------------------------
# Mutation requests
# ---------------------------------------------------------------------------


class CreateRoost(BaseModel):
    name: str
    description: str | None = None
    secret: str | None = None
    source_type: SourceType | None = None


class UpdateRoost(BaseModel):
    name: str | None = None
    description: str | None = None
    secret: str | None = None
    source_type: SourceType | None = None
    is_active: bool | None = None


class CreateDestination(BaseModel):
    destination_type: DestinationType
    name: str | None = None
    config: dict[str, Any] | None = None
    filter_expression: str | None = None
    template: str | None = None
    retry_max: int | None = None
    retry_delay_ms: int | None = None
    retry_multiplier: float | None = None


class UpdateDestination(BaseModel):
    name: str | None = None
    config: dict[str, Any] | None = None
    filter_expression: str | None = None
    template: str | None = None
    transform_type: str | None = None
    transform_code: str | None = None


class UpdateProfileRequest(BaseModel):
    name: str | None = None
    email: str | None = None


class CreateApiKeyRequest(BaseModel):
    name: str | None = None


class UpdateApiKeyRequest(BaseModel):
    name: str


# ---------------------------------------------------------------------------
# Responses
# ---------------------------------------------------------------------------


class ReplayResponse(BaseModel):
    replayed: bool
    pigeon_id: str
    delivery_attempts: int


class PauseResponse(BaseModel):
    is_paused: bool


class SendResponse(BaseModel):
    id: str
    status: str
    destinations: int


# ---------------------------------------------------------------------------
# Pagination
# ---------------------------------------------------------------------------


class PaginatedPigeons(BaseModel):
    data: list[Pigeon]
    next_cursor: str | None = None
    has_more: bool


class PaginatedDeliveryAttempts(BaseModel):
    data: list[DeliveryAttempt]
    next_cursor: str | None = None
    has_more: bool


# ---------------------------------------------------------------------------
# Templates
# ---------------------------------------------------------------------------


class Template(BaseModel):
    id: str
    name: str
    description: str
    body: str
    created_at: str
    updated_at: str


class CreateTemplate(BaseModel):
    name: str
    description: str | None = None
    body: str | None = None


class UpdateTemplate(BaseModel):
    name: str | None = None
    description: str | None = None
    body: str | None = None


class PreviewTemplateRequest(BaseModel):
    body: str
    pigeon_id: str


class PreviewTemplateResponse(BaseModel):
    rendered: str



"""Tests for SSE event streaming."""

from __future__ import annotations


def test_parse_sse_lines() -> None:
    """Verify the data: prefix stripping logic used by the event stream."""
    lines = [
        'data: {"id":"p1","roost_id":"r1"}',
        "",
        'data: {"id":"p2","roost_id":"r1"}',
        ": keepalive",
        "event: message",
    ]
    results: list[str] = []
    for line in lines:
        if line.startswith("data:"):
            results.append(line[5:].strip())

    assert len(results) == 2
    assert '"p1"' in results[0]
    assert '"p2"' in results[1]


def test_empty_data_line() -> None:
    """A data: line with no payload should yield an empty string."""
    line = "data:"
    result = line[5:].strip()
    assert result == ""


def test_data_with_extra_whitespace() -> None:
    """Whitespace after data: should be stripped."""
    line = 'data:   {"id":"p1"}  '
    result = line[5:].strip()
    assert result == '{"id":"p1"}'



"""Asynchronous client for the pgns API."""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import Any
from urllib.parse import quote

import httpx

from pgns.sdk._client import _auth_headers, _handle_response
from pgns.sdk.errors import PigeonsAuthError, PigeonsError
from pgns.sdk.models import (
    ApiKeyCreatedResponse,
    ApiKeyResponse,
    AuthTokens,
    CreateApiKeyRequest,
    CreateDestination,
    CreateRoost,
    CreateTemplate,
    Destination,
    PaginatedDeliveryAttempts,
    PaginatedPigeons,
    PauseResponse,
    Pigeon,
    PreviewTemplateRequest,
    PreviewTemplateResponse,
    ReplayResponse,
    Roost,
    SendResponse,
    Template,
    UpdateApiKeyRequest,
    UpdateDestination,
    UpdateProfileRequest,
    UpdateRoost,
    UpdateTemplate,
    User,
)


class AsyncPigeonsClient:
    """Asynchronous client for the pgns webhook relay API.

    Async mirror of :class:`~pgns.sdk.client.PigeonsClient`.

    Example::

        async with httpx.AsyncClient() as http:
            client = AsyncPigeonsClient(
                "https://api.pgns.io",
                api_key="pk_live_...",
                http_client=http,
            )
            roosts = await client.list_roosts()
    """

    def __init__(
        self,
        base_url: str,
        *,
        api_key: str | None = None,
        access_token: str | None = None,
        on_token_refresh: Callable[[AuthTokens], None] | None = None,
        http_client: httpx.AsyncClient | None = None,
    ) -> None:
        self._base_url = base_url.rstrip("/")
        self._api_key = api_key
        self._access_token = access_token
        self._on_token_refresh = on_token_refresh
        self._http = http_client or httpx.AsyncClient()
        self._refresh_lock = asyncio.Lock()

    # -- Credential setters ---------------------------------------------------

    def set_access_token(self, token: str) -> None:
        self._access_token = token

    def set_api_key(self, key: str) -> None:
        self._api_key = key

    # -- Internal helpers -----------------------------------------------------

    def _headers(self) -> dict[str, str]:
        return {
            "Content-Type": "application/json",
            **_auth_headers(self._api_key, self._access_token),
        }

    async def _refresh_token(self) -> AuthTokens:
        async with self._refresh_lock:
            response = await self._http.post(
                f"{self._base_url}/v1/auth/refresh",
                headers={"Content-Type": "application/json"},
            )
            data = _handle_response(response)
            tokens = AuthTokens.model_validate(data)
            self._access_token = tokens.access_token
            if self._on_token_refresh:
                self._on_token_refresh(tokens)
            return tokens

    async def _request(self, method: str, path: str, *, json: Any = None) -> Any:
        response = await self._http.request(
            method,
            f"{self._base_url}{path}",
            headers=self._headers(),
            json=json,
        )
        if response.status_code == 401 and not self._api_key:
            try:
                tokens = await self._refresh_token()
                headers = {
                    "Content-Type": "application/json",
                    "Authorization": f"Bearer {tokens.access_token}",
                }
                response = await self._http.request(
                    method,
                    f"{self._base_url}{path}",
                    headers=headers,
                    json=json,
                )
                return _handle_response(response)
            except PigeonsError:
                raise
            except Exception as exc:
                self._access_token = None
                raise PigeonsAuthError() from exc
        return _handle_response(response)

    async def _unauth_request(self, method: str, path: str, *, json: Any = None) -> Any:
        response = await self._http.request(
            method,
            f"{self._base_url}{path}",
            headers={"Content-Type": "application/json"},
            json=json,
        )
        return _handle_response(response)

    # -- Auth -----------------------------------------------------------------

    async def refresh(self) -> AuthTokens:
        """Refresh the access token using the httpOnly refresh cookie."""
        raw = await self._unauth_request("POST", "/v1/auth/refresh")
        tokens = AuthTokens.model_validate(raw)
        self._access_token = tokens.access_token
        if self._on_token_refresh:
            self._on_token_refresh(tokens)
        return tokens

    async def logout(self) -> None:
        """Revoke the refresh token and clear stored credentials."""
        await self._request("POST", "/v1/auth/logout")
        self._access_token = None

    # -- Roosts ---------------------------------------------------------------

    async def list_roosts(self) -> list[Roost]:
        data = await self._request("GET", "/v1/roosts")
        return [Roost.model_validate(r) for r in data]

    async def get_roost(self, roost_id: str) -> Roost:
        data = await self._request("GET", f"/v1/roosts/{quote(roost_id, safe='')}")
        return Roost.model_validate(data)

    async def create_roost(self, data: CreateRoost) -> Roost:
        raw = await self._request("POST", "/v1/roosts", json=data.model_dump(exclude_none=True))
        return Roost.model_validate(raw)

    async def update_roost(self, roost_id: str, data: UpdateRoost) -> Roost:
        raw = await self._request(
            "PATCH",
            f"/v1/roosts/{quote(roost_id, safe='')}",
            json=data.model_dump(exclude_unset=True),
        )
        return Roost.model_validate(raw)

    async def delete_roost(self, roost_id: str) -> None:
        await self._request("DELETE", f"/v1/roosts/{quote(roost_id, safe='')}")

    # -- Pigeons --------------------------------------------------------------

    async def list_pigeons(
        self,
        *,
        roost_id: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> PaginatedPigeons:
        params: list[str] = []
        if roost_id:
            params.append(f"roost_id={quote(roost_id, safe='')}")
        if limit is not None:
            params.append(f"limit={limit}")
        if cursor:
            params.append(f"cursor={quote(cursor, safe='')}")
        qs = f"?{'&'.join(params)}" if params else ""
        data = await self._request("GET", f"/v1/pigeons{qs}")
        return PaginatedPigeons.model_validate(data)

    async def get_pigeon(self, pigeon_id: str) -> Pigeon:
        data = await self._request("GET", f"/v1/pigeons/{quote(pigeon_id, safe='')}")
        return Pigeon.model_validate(data)

    async def get_pigeon_deliveries(
        self,
        pigeon_id: str,
        *,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> PaginatedDeliveryAttempts:
        params: list[str] = []
        if limit is not None:
            params.append(f"limit={limit}")
        if cursor:
            params.append(f"cursor={quote(cursor, safe='')}")
        qs = f"?{'&'.join(params)}" if params else ""
        data = await self._request("GET", f"/v1/pigeons/{quote(pigeon_id, safe='')}/deliveries{qs}")
        return PaginatedDeliveryAttempts.model_validate(data)

    async def replay_pigeon(self, pigeon_id: str) -> ReplayResponse:
        data = await self._request("POST", f"/v1/pigeons/{quote(pigeon_id, safe='')}/replay")
        return ReplayResponse.model_validate(data)

    # -- Destinations ---------------------------------------------------------

    async def list_destinations(self, roost_id: str) -> list[Destination]:
        data = await self._request("GET", f"/v1/roosts/{quote(roost_id, safe='')}/destinations")
        return [Destination.model_validate(d) for d in data]

    async def get_destination(self, destination_id: str) -> Destination:
        data = await self._request("GET", f"/v1/destinations/{quote(destination_id, safe='')}")
        return Destination.model_validate(data)

    async def create_destination(self, roost_id: str, data: CreateDestination) -> Destination:
        raw = await self._request(
            "POST",
            f"/v1/roosts/{quote(roost_id, safe='')}/destinations",
            json=data.model_dump(exclude_none=True),
        )
        return Destination.model_validate(raw)

    async def pause_destination(self, destination_id: str, is_paused: bool) -> PauseResponse:
        data = await self._request(
            "PATCH",
            f"/v1/destinations/{quote(destination_id, safe='')}/pause",
            json={"is_paused": is_paused},
        )
        return PauseResponse.model_validate(data)

    async def update_destination(self, destination_id: str, data: UpdateDestination) -> Destination:
        raw = await self._request(
            "PATCH",
            f"/v1/destinations/{quote(destination_id, safe='')}",
            json=data.model_dump(exclude_unset=True),
        )
        return Destination.model_validate(raw)

    async def delete_destination(self, destination_id: str) -> None:
        await self._request("DELETE", f"/v1/destinations/{quote(destination_id, safe='')}")

    # -- API Keys -------------------------------------------------------------

    async def list_api_keys(self) -> list[ApiKeyResponse]:
        data = await self._request("GET", "/v1/api-keys")
        return [ApiKeyResponse.model_validate(k) for k in data]

    async def get_api_key(self, key_id: str) -> ApiKeyResponse:
        data = await self._request("GET", f"/v1/api-keys/{quote(key_id, safe='')}")
        return ApiKeyResponse.model_validate(data)

    async def create_api_key(
        self, data: CreateApiKeyRequest | None = None
    ) -> ApiKeyCreatedResponse:
        body = data.model_dump(exclude_none=True) if data else {}
        raw = await self._request("POST", "/v1/api-keys", json=body)
        return ApiKeyCreatedResponse.model_validate(raw)

    async def update_api_key(self, key_id: str, data: UpdateApiKeyRequest) -> ApiKeyResponse:
        raw = await self._request(
            "PATCH",
            f"/v1/api-keys/{quote(key_id, safe='')}",
            json=data.model_dump(),
        )
        return ApiKeyResponse.model_validate(raw)

    async def delete_api_key(self, key_id: str) -> None:
        await self._request("DELETE", f"/v1/api-keys/{quote(key_id, safe='')}")

    # -- Templates ------------------------------------------------------------

    async def list_templates(self) -> list[Template]:
        data = await self._request("GET", "/v1/templates")
        return [Template.model_validate(t) for t in data]

    async def get_template(self, template_id: str) -> Template:
        data = await self._request("GET", f"/v1/templates/{quote(template_id, safe='')}")
        return Template.model_validate(data)

    async def create_template(self, data: CreateTemplate) -> Template:
        raw = await self._request("POST", "/v1/templates", json=data.model_dump(exclude_none=True))
        return Template.model_validate(raw)

    async def update_template(self, template_id: str, data: UpdateTemplate) -> Template:
        raw = await self._request(
            "PATCH",
            f"/v1/templates/{quote(template_id, safe='')}",
            json=data.model_dump(exclude_unset=True),
        )
        return Template.model_validate(raw)

    async def delete_template(self, template_id: str) -> None:
        await self._request("DELETE", f"/v1/templates/{quote(template_id, safe='')}")

    async def preview_template(self, data: PreviewTemplateRequest) -> PreviewTemplateResponse:
        raw = await self._request("POST", "/v1/templates/preview", json=data.model_dump())
        return PreviewTemplateResponse.model_validate(raw)

    # -- User -----------------------------------------------------------------

    async def get_me(self) -> User:
        data = await self._request("GET", "/v1/me")
        return User.model_validate(data)

    async def update_me(self, data: UpdateProfileRequest) -> User:
        raw = await self._request("PATCH", "/v1/me", json=data.model_dump(exclude_unset=True))
        return User.model_validate(raw)

    # -- Send (Inbound Webhook) -----------------------------------------------

    async def send(
        self,
        roost_id: str,
        *,
        event_type: str,
        payload: Any,
        signing_secret: str,
    ) -> SendResponse:
        """Send a signed webhook to a roost.

        Computes dual signatures: legacy ``X-Pigeon-Signature`` (hex) and
        Standard Webhooks ``webhook-signature`` (base64). Both header sets
        are sent for backward compatibility.
        """
        import base64
        import hashlib
        import hmac as hmac_mod
        import json
        import time
        import uuid

        body = json.dumps(payload, separators=(",", ":"))
        timestamp = str(int(time.time()))
        msg_id = f"msg_{uuid.uuid4()}"

        # Decode signing key
        if signing_secret.startswith("whsec_"):
            key_bytes = base64.b64decode(signing_secret[6:])
        else:
            key_bytes = signing_secret.encode()

        # Legacy signature: HMAC-SHA256("{timestamp}.{body}") → hex
        legacy_sig = hmac_mod.new(
            key_bytes,
            f"{timestamp}.{body}".encode(),
            hashlib.sha256,
        ).hexdigest()

        # Standard Webhooks signature: HMAC-SHA256("{msg_id}.{timestamp}.{body}") → base64
        std_sig = base64.b64encode(
            hmac_mod.new(
                key_bytes,
                f"{msg_id}.{timestamp}.{body}".encode(),
                hashlib.sha256,
            ).digest()
        ).decode()

        response = await self._http.post(
            f"{self._base_url}/r/{quote(roost_id, safe='')}",
            headers={
                "Content-Type": "application/json",
                "X-Pigeon-Signature": f"sha256={legacy_sig}",
                "X-Pigeon-Timestamp": timestamp,
                "X-Pigeon-Event-Type": event_type,
                "webhook-id": msg_id,
                "webhook-timestamp": timestamp,
                "webhook-signature": f"v1,{std_sig}",
            },
            content=body,
        )
        data = _handle_response(response)
        return SendResponse.model_validate(data)



"""Errors raised by the pgns SDK."""

from __future__ import annotations


class PigeonsError(Exception):
    """Error returned by the pgns API."""

    def __init__(self, message: str, status: int, code: str | None = None) -> None:
        super().__init__(message)
        self.status = status
        self.code = code

    def __repr__(self) -> str:
        return f"PigeonsError({self.args[0]!r}, status={self.status}, code={self.code!r})"

    def is_not_found(self) -> bool:
        return self.status == 404

    def is_unauthorized(self) -> bool:
        return self.status == 401


class PigeonsAuthError(PigeonsError):
    """Raised when authentication fails and cannot be refreshed."""

    def __init__(self, message: str = "Session expired") -> None:
        super().__init__(message, 401)


class WebhookVerificationError(PigeonsError):
    """Raised when webhook signature verification fails."""

    def __init__(self, message: str, code: str) -> None:
        super().__init__(message, 400, code=code)



"""Webhook signature verification for pgns consumers."""

from __future__ import annotations

import base64
import hashlib
import hmac as hmac_mod
import json
import time
from collections.abc import Callable
from typing import Any

from pgns.sdk.errors import WebhookVerificationError

type _HeaderGetter = Callable[[str], str | None]

DEFAULT_TOLERANCE_SECONDS = 300


class Webhook:
    """Verify incoming webhook signatures from pgns."""

    def __init__(
        self, secret: str, *, tolerance_in_seconds: int = DEFAULT_TOLERANCE_SECONDS
    ) -> None:
        self._secret = secret
        self._tolerance = tolerance_in_seconds
        self._key_bytes = self._decode_secret(secret)

    def verify(self, body: str | bytes, headers: dict[str, str]) -> Any:
        """Verify the webhook signature and return the parsed JSON payload.

        Raises ``WebhookVerificationError`` on failure.
        """
        body_str = body.decode() if isinstance(body, bytes) else body

        get = self._make_getter(headers)

        webhook_sig = get("webhook-signature")
        if webhook_sig is not None:
            return self._verify_standard_webhooks(body_str, get, webhook_sig)

        pigeon_sig = get("x-pigeon-signature")
        if pigeon_sig is not None:
            return self._verify_legacy(body_str, get, pigeon_sig)

        raise WebhookVerificationError("No signature header found", "MISSING_HEADER")

    def _verify_standard_webhooks(
        self,
        body: str,
        get: _HeaderGetter,
        sig_header: str,
    ) -> Any:
        if not sig_header.startswith("v1,"):
            raise WebhookVerificationError(
                "Invalid webhook-signature format: missing v1, prefix",
                "INVALID_FORMAT",
            )
        b64_sig = sig_header[3:]

        msg_id = get("webhook-id")
        if msg_id is None:
            raise WebhookVerificationError("Missing webhook-id header", "MISSING_HEADER")

        timestamp = get("webhook-timestamp")
        if timestamp is None:
            raise WebhookVerificationError("Missing webhook-timestamp header", "MISSING_HEADER")

        self._check_timestamp(timestamp)

        signed_payload = f"{msg_id}.{timestamp}.{body}".encode()
        mac = hmac_mod.new(self._key_bytes, signed_payload, hashlib.sha256).digest()
        expected = base64.b64encode(mac).decode()

        if not hmac_mod.compare_digest(expected, b64_sig):
            raise WebhookVerificationError("Signature mismatch", "SIGNATURE_MISMATCH")

        return json.loads(body)

    def _verify_legacy(
        self,
        body: str,
        get: _HeaderGetter,
        sig_header: str,
    ) -> Any:
        if not sig_header.startswith("sha256="):
            raise WebhookVerificationError(
                "Invalid X-Pigeon-Signature format: missing sha256= prefix",
                "INVALID_FORMAT",
            )
        hex_digest = sig_header[7:]

        timestamp = get("x-pigeon-timestamp")
        if timestamp is not None:
            self._check_timestamp(timestamp)

        payload = f"{timestamp}.{body}" if timestamp else body
        signed_payload = payload.encode()
        mac = hmac_mod.new(self._key_bytes, signed_payload, hashlib.sha256).digest()
        expected = mac.hex()

        if not hmac_mod.compare_digest(expected, hex_digest):
            raise WebhookVerificationError("Signature mismatch", "SIGNATURE_MISMATCH")

        return json.loads(body)

    def _check_timestamp(self, ts: str) -> None:
        try:
            timestamp = int(ts)
        except ValueError:
            raise WebhookVerificationError("Invalid timestamp", "INVALID_FORMAT") from None
        now = int(time.time())
        if abs(now - timestamp) > self._tolerance:
            raise WebhookVerificationError("Timestamp outside tolerance", "TIMESTAMP_EXPIRED")

    @staticmethod
    def _decode_secret(secret: str) -> bytes:
        if secret.startswith("whsec_"):
            return base64.b64decode(secret[6:])
        if len(secret) == 64 and all(c in "0123456789abcdefABCDEF" for c in secret):
            return bytes.fromhex(secret)
        return secret.encode()

    @staticmethod
    def _make_getter(headers: dict[str, str]) -> _HeaderGetter:
        lower_map = {k.lower(): v for k, v in headers.items()}
        return lambda name: lower_map.get(name.lower())



"""Tests for the synchronous PigeonsClient."""

from __future__ import annotations

import json
from typing import Any

import httpx
import pytest

from pgns.sdk.client import PigeonsClient
from pgns.sdk.errors import PigeonsError
from pgns.sdk.models import (
    CreateRoost,
    UpdateRoost,
)
from pgns.sdk.tests.conftest import (
    API_KEY,
    BASE_URL,
    SAMPLE_API_KEY,
    SAMPLE_DESTINATION,
    SAMPLE_PIGEON,
    SAMPLE_ROOST,
    SAMPLE_TEMPLATE,
    SAMPLE_USER,
    make_client,
)


class TestAuth:
    def test_auth_header_api_key(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            assert request.headers["authorization"] == f"Bearer {API_KEY}"
            return httpx.Response(200, json=[])

        client = make_client(handler)
        client.list_roosts()

    def test_auth_header_access_token(self) -> None:
        token = "eyJhbG.test.token"

        def handler(request: httpx.Request) -> httpx.Response:
            assert request.headers["authorization"] == f"Bearer {token}"
            return httpx.Response(200, json=[])

        transport = httpx.MockTransport(handler)
        http = httpx.Client(transport=transport)
        client = PigeonsClient(BASE_URL, access_token=token, http_client=http)
        client.list_roosts()

    def test_logout(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            assert request.url.path == "/v1/auth/logout"
            return httpx.Response(204)

        client = make_client(handler)
        client.logout()

    def test_401_refresh_retry(self) -> None:
        """On 401 with JWT auth, the client should refresh and retry."""
        call_count = 0
        tokens_data = {"access_token": "refreshed", "token_type": "Bearer", "expires_in": 3600}

        def handler(request: httpx.Request) -> httpx.Response:
            nonlocal call_count
            call_count += 1
            if request.url.path == "/v1/auth/refresh":
                return httpx.Response(200, json=tokens_data)
            if call_count == 1:
                return httpx.Response(401, json={"error": "token expired"})
            return httpx.Response(200, json=[SAMPLE_ROOST])

        transport = httpx.MockTransport(handler)
        http = httpx.Client(transport=transport)
        client = PigeonsClient(BASE_URL, access_token="expired", http_client=http)
        roosts = client.list_roosts()
        assert len(roosts) == 1
        assert call_count == 3  # original + refresh + retry


class TestErrorHandling:
    def test_404_raises(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            return httpx.Response(404, json={"error": "not found"})

        client = make_client(handler)
        with pytest.raises(PigeonsError) as exc_info:
            client.get_roost("nonexistent")
        assert exc_info.value.status == 404
        assert exc_info.value.is_not_found()

    def test_500_raises(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            return httpx.Response(500, json={"error": "internal error"})

        client = make_client(handler)
        with pytest.raises(PigeonsError) as exc_info:
            client.list_roosts()
        assert exc_info.value.status == 500

    def test_non_json_error(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            return httpx.Response(502, content=b"Bad Gateway")

        client = make_client(handler)
        with pytest.raises(PigeonsError):
            client.list_roosts()


class TestRoosts:
    def test_list_roosts(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            assert request.url.path == "/v1/roosts"
            return httpx.Response(200, json=[SAMPLE_ROOST])

        client = make_client(handler)
        roosts = client.list_roosts()
        assert len(roosts) == 1
        assert roosts[0].id == "roost_abc123"
        assert roosts[0].is_active is True

    def test_get_roost(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            assert request.url.path == "/v1/roosts/roost_abc123"
            return httpx.Response(200, json=SAMPLE_ROOST)

        client = make_client(handler)
        roost = client.get_roost("roost_abc123")
        assert roost.name == "Test Roost"

    def test_create_roost(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            body = json.loads(request.content)
            assert body["name"] == "New Roost"
            assert "description" not in body  # exclude_none
            return httpx.Response(201, json={**SAMPLE_ROOST, "name": "New Roost"})

        client = make_client(handler)
        roost = client.create_roost(CreateRoost(name="New Roost"))
        assert roost.name == "New Roost"

    def test_update_roost(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            body = json.loads(request.content)
            assert body == {"name": "Updated"}  # exclude_unset
            return httpx.Response(200, json={**SAMPLE_ROOST, "name": "Updated"})

        client = make_client(handler)
        roost = client.update_roost("roost_abc123", UpdateRoost(name="Updated"))
        assert roost.name == "Updated"

    def test_delete_roost(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            assert request.method == "DELETE"
            return httpx.Response(204)

        client = make_client(handler)
        client.delete_roost("roost_abc123")


class TestPigeons:
    def test_list_pigeons(self) -> None:
        paginated = {"data": [SAMPLE_PIGEON], "next_cursor": None, "has_more": False}

        def handler(request: httpx.Request) -> httpx.Response:
            assert request.url.path == "/v1/pigeons"
            return httpx.Response(200, json=paginated)

        client = make_client(handler)
        result = client.list_pigeons()
        assert len(result.data) == 1
        assert result.has_more is False

    def test_list_pigeons_with_filters(self) -> None:
        paginated: dict[str, Any] = {"data": [], "next_cursor": None, "has_more": False}

        def handler(request: httpx.Request) -> httpx.Response:
            assert "roost_id=roost_abc123" in str(request.url)
            assert "limit=10" in str(request.url)
            return httpx.Response(200, json=paginated)

        client = make_client(handler)
        client.list_pigeons(roost_id="roost_abc123", limit=10)

    def test_replay_pigeon(self) -> None:
        replay = {"replayed": True, "pigeon_id": "pgn_abc123", "delivery_attempts": 2}

        def handler(request: httpx.Request) -> httpx.Response:
            assert request.method == "POST"
            assert request.url.path == "/v1/pigeons/pgn_abc123/replay"
            return httpx.Response(200, json=replay)

        client = make_client(handler)
        result = client.replay_pigeon("pgn_abc123")
        assert result.replayed is True
        assert result.delivery_attempts == 2


class TestDestinations:
    def test_list_destinations(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            assert request.url.path == "/v1/roosts/roost_abc123/destinations"
            return httpx.Response(200, json=[SAMPLE_DESTINATION])

        client = make_client(handler)
        dests = client.list_destinations("roost_abc123")
        assert len(dests) == 1
        assert dests[0].destination_type.value == "url"

    def test_pause_destination(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            body = json.loads(request.content)
            assert body["is_paused"] is True
            return httpx.Response(200, json={"is_paused": True})

        client = make_client(handler)
        result = client.pause_destination("dest_abc123", True)
        assert result.is_paused is True

    def test_delete_destination(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            assert request.method == "DELETE"
            return httpx.Response(204)

        client = make_client(handler)
        client.delete_destination("dest_abc123")


class TestApiKeys:
    def test_list_api_keys(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            return httpx.Response(200, json=[SAMPLE_API_KEY])

        client = make_client(handler)
        keys = client.list_api_keys()
        assert len(keys) == 1
        assert keys[0].key_prefix == "pk_live_test1234"

    def test_create_api_key(self) -> None:
        created = {**SAMPLE_API_KEY, "key": "pk_live_fullkeyvalue1234567890abcdef12345678"}

        def handler(request: httpx.Request) -> httpx.Response:
            return httpx.Response(201, json=created)

        client = make_client(handler)
        result = client.create_api_key()
        assert result.key.startswith("pk_live_")


class TestTemplates:
    def test_list_templates(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            return httpx.Response(200, json=[SAMPLE_TEMPLATE])

        client = make_client(handler)
        templates = client.list_templates()
        assert len(templates) == 1
        assert templates[0].name == "Test Template"

    def test_preview_template(self) -> None:
        from pgns.sdk.models import PreviewTemplateRequest

        def handler(request: httpx.Request) -> httpx.Response:
            body = json.loads(request.content)
            assert body["body"] == "{{ body }}"
            assert body["pigeon_id"] == "pgn_abc123"
            return httpx.Response(200, json={"rendered": "hello world"})

        client = make_client(handler)
        result = client.preview_template(
            PreviewTemplateRequest(body="{{ body }}", pigeon_id="pgn_abc123")
        )
        assert result.rendered == "hello world"


class TestSend:
    def test_send_signed_webhook(self) -> None:
        """send() computes HMAC-SHA256 and sets correct headers."""
        import hashlib
        import hmac as hmac_mod

        captured: dict[str, Any] = {}

        def handler(request: httpx.Request) -> httpx.Response:
            captured["url"] = str(request.url)
            captured["headers"] = dict(request.headers)
            captured["body"] = request.content.decode()
            return httpx.Response(
                200,
                json={"id": "pgn_test123", "status": "received", "destinations": 1},
            )

        client = make_client(handler)
        result = client.send(
            "rst_abc",
            event_type="order.created",
            payload={"order_id": "123"},
            signing_secret="test-secret",
        )

        assert result.id == "pgn_test123"
        assert result.status == "received"
        assert result.destinations == 1
        assert captured["url"].endswith("/r/rst_abc")
        assert captured["headers"]["x-pigeon-event-type"] == "order.created"
        assert captured["headers"]["x-pigeon-signature"].startswith("sha256=")

        # Verify the legacy HMAC is correct
        ts = captured["headers"]["x-pigeon-timestamp"]
        body = captured["body"]
        expected_sig = hmac_mod.new(
            b"test-secret",
            f"{ts}.{body}".encode(),
            hashlib.sha256,
        ).hexdigest()
        assert captured["headers"]["x-pigeon-signature"] == f"sha256={expected_sig}"

        # Verify Standard Webhooks headers
        import base64
        import re

        msg_id = captured["headers"]["webhook-id"]
        assert re.match(r"^msg_[0-9a-f-]{36}$", msg_id)
        assert captured["headers"]["webhook-timestamp"] == ts
        wh_sig = captured["headers"]["webhook-signature"]
        assert wh_sig.startswith("v1,")

        # Verify Standard Webhooks HMAC
        expected_std = base64.b64encode(
            hmac_mod.new(
                b"test-secret",
                f"{msg_id}.{ts}.{body}".encode(),
                hashlib.sha256,
            ).digest()
        ).decode()
        assert wh_sig == f"v1,{expected_std}"


class TestUserAndStats:
    def test_get_me(self) -> None:
        def handler(request: httpx.Request) -> httpx.Response:
            assert request.url.path == "/v1/me"
            return httpx.Response(200, json=SAMPLE_USER)

        client = make_client(handler)
        user = client.get_me()
        assert user.email == "test@example.com"



from __future__ import annotations

import base64
import hashlib
import hmac as hmac_mod
import time

import pytest

from pgns.sdk.errors import WebhookVerificationError
from pgns.sdk.webhook import Webhook

SECRET = "test-secret"


def _now_ts() -> str:
    return str(int(time.time()))


def _standard_headers(
    secret: str,
    body: str,
    *,
    timestamp: str | None = None,
    msg_id: str = "msg_test-id",
) -> dict[str, str]:
    ts = timestamp or _now_ts()
    signed_payload = f"{msg_id}.{ts}.{body}".encode()
    mac = hmac_mod.new(secret.encode(), signed_payload, hashlib.sha256).digest()
    b64 = base64.b64encode(mac).decode()
    return {
        "webhook-id": msg_id,
        "webhook-timestamp": ts,
        "webhook-signature": f"v1,{b64}",
    }


def _legacy_headers(
    secret: str,
    body: str,
    *,
    timestamp: str | None = None,
) -> dict[str, str]:
    ts = timestamp or _now_ts()
    payload = f"{ts}.{body}"
    mac = hmac_mod.new(secret.encode(), payload.encode(), hashlib.sha256).digest()
    return {
        "X-Pigeon-Signature": f"sha256={mac.hex()}",
        "X-Pigeon-Timestamp": ts,
    }


class TestStandardWebhooks:
    def test_verify_valid(self) -> None:
        body = '{"event":"test"}'
        headers = _standard_headers(SECRET, body)
        wh = Webhook(SECRET)
        result = wh.verify(body, headers)
        assert result == {"event": "test"}

    def test_reject_tampered_body(self) -> None:
        body = '{"event":"test"}'
        headers = _standard_headers(SECRET, body)
        wh = Webhook(SECRET)
        with pytest.raises(WebhookVerificationError, match="Signature mismatch"):
            wh.verify('{"event":"tampered"}', headers)

    def test_reject_wrong_secret(self) -> None:
        body = '{"event":"test"}'
        headers = _standard_headers(SECRET, body)
        wh = Webhook("wrong-secret")
        with pytest.raises(WebhookVerificationError, match="Signature mismatch"):
            wh.verify(body, headers)

    def test_reject_expired_timestamp(self) -> None:
        body = '{"event":"test"}'
        old_ts = str(int(time.time()) - 600)
        headers = _standard_headers(SECRET, body, timestamp=old_ts)
        wh = Webhook(SECRET)
        with pytest.raises(WebhookVerificationError, match="Timestamp outside tolerance"):
            wh.verify(body, headers)

    def test_custom_tolerance(self) -> None:
        body = '{"event":"test"}'
        old_ts = str(int(time.time()) - 600)
        headers = _standard_headers(SECRET, body, timestamp=old_ts)
        wh = Webhook(SECRET, tolerance_in_seconds=700)
        result = wh.verify(body, headers)
        assert result == {"event": "test"}

    def test_reject_missing_webhook_id(self) -> None:
        body = '{"event":"test"}'
        headers = _standard_headers(SECRET, body)
        del headers["webhook-id"]
        wh = Webhook(SECRET)
        with pytest.raises(WebhookVerificationError, match="Missing webhook-id"):
            wh.verify(body, headers)

    def test_reject_missing_webhook_timestamp(self) -> None:
        body = '{"event":"test"}'
        headers = _standard_headers(SECRET, body)
        del headers["webhook-timestamp"]
        wh = Webhook(SECRET)
        with pytest.raises(WebhookVerificationError, match="Missing webhook-timestamp"):
            wh.verify(body, headers)

    def test_reject_invalid_format(self) -> None:
        body = '{"event":"test"}'
        wh = Webhook(SECRET)
        with pytest.raises(WebhookVerificationError, match="missing v1, prefix"):
            wh.verify(
                body,
                {
                    "webhook-id": "msg_1",
                    "webhook-timestamp": _now_ts(),
                    "webhook-signature": "bad-format",
                },
            )


class TestLegacyFormat:
    def test_verify_valid(self) -> None:
        body = '{"event":"test"}'
        headers = _legacy_headers(SECRET, body)
        wh = Webhook(SECRET)
        result = wh.verify(body, headers)
        assert result == {"event": "test"}

    def test_reject_tampered_body(self) -> None:
        body = '{"event":"test"}'
        headers = _legacy_headers(SECRET, body)
        wh = Webhook(SECRET)
        with pytest.raises(WebhookVerificationError, match="Signature mismatch"):
            wh.verify('{"event":"tampered"}', headers)

    def test_reject_expired_timestamp(self) -> None:
        body = '{"event":"test"}'
        old_ts = str(int(time.time()) - 600)
        headers = _legacy_headers(SECRET, body, timestamp=old_ts)
        wh = Webhook(SECRET)
        with pytest.raises(WebhookVerificationError, match="Timestamp outside tolerance"):
            wh.verify(body, headers)

    def test_reject_invalid_format(self) -> None:
        body = '{"event":"test"}'
        wh = Webhook(SECRET)
        with pytest.raises(WebhookVerificationError, match="missing sha256= prefix"):
            wh.verify(body, {"X-Pigeon-Signature": "bad-format"})


class TestFormatDetection:
    def test_prefers_standard_when_both_present(self) -> None:
        body = '{"event":"test"}'
        std = _standard_headers(SECRET, body)
        lgc = _legacy_headers(SECRET, body)
        combined = {**lgc, **std}
        wh = Webhook(SECRET)
        result = wh.verify(body, combined)
        assert result == {"event": "test"}

    def test_missing_header_raises(self) -> None:
        wh = Webhook(SECRET)
        with pytest.raises(WebhookVerificationError, match="No signature header found"):
            wh.verify("{}", {})


class TestHeaderCaseInsensitivity:
    def test_uppercase_headers(self) -> None:
        body = '{"event":"test"}'
        headers = _standard_headers(SECRET, body)
        upper = {k.upper(): v for k, v in headers.items()}
        wh = Webhook(SECRET)
        result = wh.verify(body, upper)
        assert result == {"event": "test"}


class TestKeyDecoding:
    def test_whsec_prefix(self) -> None:
        import os

        raw_key = os.urandom(32)
        b64 = base64.b64encode(raw_key).decode()
        whsec_secret = f"whsec_{b64}"

        body = '{"event":"test"}'
        ts = _now_ts()
        msg_id = "msg_whsec-test"
        signed_payload = f"{msg_id}.{ts}.{body}".encode()
        mac = hmac_mod.new(raw_key, signed_payload, hashlib.sha256).digest()
        sig = base64.b64encode(mac).decode()

        headers = {
            "webhook-id": msg_id,
            "webhook-timestamp": ts,
            "webhook-signature": f"v1,{sig}",
        }

        wh = Webhook(whsec_secret)
        result = wh.verify(body, headers)
        assert result == {"event": "test"}

    def test_hex_key(self) -> None:
        import os

        raw_key = os.urandom(32)
        hex_secret = raw_key.hex()

        body = '{"event":"test"}'
        ts = _now_ts()
        msg_id = "msg_hex-test"
        signed_payload = f"{msg_id}.{ts}.{body}".encode()
        mac = hmac_mod.new(raw_key, signed_payload, hashlib.sha256).digest()
        sig = base64.b64encode(mac).decode()

        headers = {
            "webhook-id": msg_id,
            "webhook-timestamp": ts,
            "webhook-signature": f"v1,{sig}",
        }

        wh = Webhook(hex_secret)
        result = wh.verify(body, headers)
        assert result == {"event": "test"}

    def test_bytes_body(self) -> None:
        body = b'{"event":"test"}'
        headers = _standard_headers(SECRET, body.decode())
        wh = Webhook(SECRET)
        result = wh.verify(body, headers)
        assert result == {"event": "test"}



"""Tests for the asynchronous AsyncPigeonsClient."""

from __future__ import annotations

import json

import httpx
import pytest

from pgns.sdk.async_client import AsyncPigeonsClient
from pgns.sdk.errors import PigeonsError
from pgns.sdk.models import CreateRoost, UpdateRoost
from pgns.sdk.tests.conftest import (
    API_KEY,
    BASE_URL,
    SAMPLE_PIGEON,
    SAMPLE_ROOST,
    SAMPLE_USER,
    make_async_client,
)


@pytest.mark.asyncio
async def test_list_roosts() -> None:
    def handler(request: httpx.Request) -> httpx.Response:
        assert request.headers["authorization"] == f"Bearer {API_KEY}"
        return httpx.Response(200, json=[SAMPLE_ROOST])

    client = make_async_client(handler)
    roosts = await client.list_roosts()
    assert len(roosts) == 1
    assert roosts[0].id == "roost_abc123"


@pytest.mark.asyncio
async def test_create_roost() -> None:
    def handler(request: httpx.Request) -> httpx.Response:
        body = json.loads(request.content)
        assert body["name"] == "Async Roost"
        return httpx.Response(201, json={**SAMPLE_ROOST, "name": "Async Roost"})

    client = make_async_client(handler)
    roost = await client.create_roost(CreateRoost(name="Async Roost"))
    assert roost.name == "Async Roost"


@pytest.mark.asyncio
async def test_update_roost_exclude_unset() -> None:
    def handler(request: httpx.Request) -> httpx.Response:
        body = json.loads(request.content)
        assert body == {"name": "Updated"}
        return httpx.Response(200, json={**SAMPLE_ROOST, "name": "Updated"})

    client = make_async_client(handler)
    roost = await client.update_roost("roost_abc123", UpdateRoost(name="Updated"))
    assert roost.name == "Updated"


@pytest.mark.asyncio
async def test_delete_roost_204() -> None:
    def handler(request: httpx.Request) -> httpx.Response:
        assert request.method == "DELETE"
        return httpx.Response(204)

    client = make_async_client(handler)
    await client.delete_roost("roost_abc123")


@pytest.mark.asyncio
async def test_error_handling() -> None:
    def handler(request: httpx.Request) -> httpx.Response:
        return httpx.Response(404, json={"error": "not found"})

    client = make_async_client(handler)
    with pytest.raises(PigeonsError) as exc_info:
        await client.get_roost("nonexistent")
    assert exc_info.value.status == 404


@pytest.mark.asyncio
async def test_list_pigeons_paginated() -> None:
    paginated = {"data": [SAMPLE_PIGEON], "next_cursor": "abc", "has_more": True}

    def handler(request: httpx.Request) -> httpx.Response:
        return httpx.Response(200, json=paginated)

    client = make_async_client(handler)
    result = await client.list_pigeons(roost_id="roost_abc123", limit=1)
    assert len(result.data) == 1
    assert result.has_more is True
    assert result.next_cursor == "abc"


@pytest.mark.asyncio
async def test_401_refresh_retry() -> None:
    call_count = 0
    tokens_data = {"access_token": "refreshed", "token_type": "Bearer", "expires_in": 3600}

    def handler(request: httpx.Request) -> httpx.Response:
        nonlocal call_count
        call_count += 1
        if request.url.path == "/v1/auth/refresh":
            return httpx.Response(200, json=tokens_data)
        if call_count == 1:
            return httpx.Response(401, json={"error": "token expired"})
        return httpx.Response(200, json=SAMPLE_USER)

    transport = httpx.MockTransport(handler)
    http = httpx.AsyncClient(transport=transport)
    client = AsyncPigeonsClient(BASE_URL, access_token="expired", http_client=http)
    user = await client.get_me()
    assert user.email == "test@example.com"



"""SSE event streaming for the pgns API."""

from __future__ import annotations

import asyncio
import time
from collections.abc import AsyncIterator, Iterator

import httpx

_DEFAULT_RETRY_DELAY = 3.0


def event_stream(
    base_url: str,
    *,
    token: str | None = None,
    roost_id: str | None = None,
) -> Iterator[str]:
    """Connect to the pgns SSE event stream (synchronous).

    Yields each ``data:`` line as a string. Automatically reconnects on
    failure with a 3-second delay. Break out of the iterator to disconnect.

    Args:
        base_url: Base URL of the pgns API.
        token: Bearer token (API key or JWT) for authentication.
        roost_id: Restrict the stream to a single roost.

    Yields:
        Event data strings (JSON-encoded pigeon notifications).
    """
    url = f"{base_url.rstrip('/')}/v1/events"
    if roost_id:
        url += f"?roost_id={roost_id}"
    headers: dict[str, str] = {"Accept": "text/event-stream"}
    if token:
        headers["Authorization"] = f"Bearer {token}"

    while True:
        try:
            with httpx.Client() as client:
                with client.stream("GET", url, headers=headers, timeout=None) as response:
                    response.raise_for_status()
                    buffer = ""
                    for chunk in response.iter_text():
                        buffer += chunk
                        while "\n" in buffer:
                            line, buffer = buffer.split("\n", 1)
                            if line.startswith("data:"):
                                yield line[5:].strip()
        except GeneratorExit:
            return
        except Exception:
            pass
        time.sleep(_DEFAULT_RETRY_DELAY)


async def async_event_stream(
    base_url: str,
    *,
    token: str | None = None,
    roost_id: str | None = None,
) -> AsyncIterator[str]:
    """Connect to the pgns SSE event stream (asynchronous).

    Async mirror of :func:`event_stream`.
    """
    url = f"{base_url.rstrip('/')}/v1/events"
    if roost_id:
        url += f"?roost_id={roost_id}"
    headers: dict[str, str] = {"Accept": "text/event-stream"}
    if token:
        headers["Authorization"] = f"Bearer {token}"

    while True:
        try:
            async with httpx.AsyncClient() as client:
                async with client.stream("GET", url, headers=headers, timeout=None) as response:
                    response.raise_for_status()
                    buffer = ""
                    async for chunk in response.aiter_text():
                        buffer += chunk
                        while "\n" in buffer:
                            line, buffer = buffer.split("\n", 1)
                            if line.startswith("data:"):
                                yield line[5:].strip()
        except GeneratorExit:
            return
        except Exception:
            pass
        await asyncio.sleep(_DEFAULT_RETRY_DELAY)



"""Shared helpers for sync and async clients."""

from __future__ import annotations

from typing import Any

import httpx

from pgns.sdk.errors import PigeonsError


def _auth_headers(api_key: str | None, access_token: str | None) -> dict[str, str]:
    """Build the Authorization header dict."""
    if api_key:
        return {"Authorization": f"Bearer {api_key}"}
    if access_token:
        return {"Authorization": f"Bearer {access_token}"}
    return {}


def _handle_response(response: httpx.Response) -> Any:
    """Parse an httpx response, raising PigeonsError on non-2xx."""
    if response.status_code == 204:
        return None
    if not response.is_success:
        try:
            body = response.json()
            message = body.get("error", response.reason_phrase or "Unknown error")
            code = body.get("code")
        except Exception:
            message = response.reason_phrase or "Unknown error"
            code = None
        raise PigeonsError(message, response.status_code, code=code)
    return response.json()



"""Tests for Pydantic model serialization and validation."""

from __future__ import annotations

from typing import Any

import pytest

from pgns.sdk.models import (
    AuthTokens,
    CreateDestination,
    CreateRoost,
    DeliveryAttempt,
    DeliveryStatus,
    Destination,
    DestinationType,
    PaginatedPigeons,
    Pigeon,
    Roost,
    UpdateRoost,
    User,
)


class TestEnums:
    def test_destination_type_values(self) -> None:
        assert DestinationType.url.value == "url"
        assert DestinationType.slack.value == "slack"
        assert DestinationType.discord.value == "discord"
        assert DestinationType.email.value == "email"

    def test_delivery_status_values(self) -> None:
        assert DeliveryStatus.pending.value == "pending"
        assert DeliveryStatus.delivered.value == "delivered"
        assert DeliveryStatus.failed.value == "failed"


class TestResponseModels:
    def test_roost_from_json(self) -> None:
        data = {
            "id": "r1",
            "name": "Test",
            "description": "desc",
            "secret": None,
            "is_active": True,
            "created_at": "2024-01-01T00:00:00Z",
            "updated_at": "2024-01-01T00:00:00Z",
        }
        roost = Roost.model_validate(data)
        assert roost.id == "r1"
        assert roost.secret is None
        assert roost.is_active is True

    def test_pigeon_from_json(self) -> None:
        data = {
            "id": "p1",
            "roost_id": "r1",
            "source_ip": "1.2.3.4",
            "request_method": "POST",
            "content_type": "application/json",
            "headers": {"x-foo": "bar"},
            "body_json": {"key": "value"},
            "body_raw": None,
            "request_query": None,
            "replayed_from": None,
            "delivery_status": "delivered",
            "received_at": "2024-01-01T00:00:00Z",
        }
        pigeon = Pigeon.model_validate(data)
        assert pigeon.delivery_status == DeliveryStatus.delivered
        assert pigeon.headers["x-foo"] == "bar"

    def test_destination_from_json(self) -> None:
        data = {
            "id": "d1",
            "roost_id": "r1",
            "name": "Slack Alerts",
            "destination_type": "slack",
            "config": {"webhook_url": "https://hooks.slack.com/..."},
            "filter_expression": "",
            "template": "",
            "retry_max": 5,
            "retry_delay_ms": 1000,
            "retry_multiplier": 2.0,
            "is_paused": False,
            "is_verified": True,
            "created_at": "2024-01-01T00:00:00Z",
            "updated_at": "2024-01-01T00:00:00Z",
        }
        dest = Destination.model_validate(data)
        assert dest.destination_type == DestinationType.slack

    def test_user_from_json(self) -> None:
        data = {
            "id": "u1",
            "email": "a@b.com",
            "name": "Alice",
            "plan": "pro",
            "created_at": "2024-01-01T00:00:00Z",
            "updated_at": "2024-01-01T00:00:00Z",
        }
        user = User.model_validate(data)
        assert user.plan == "pro"

    def test_auth_tokens(self) -> None:
        data = {"access_token": "tok", "token_type": "Bearer", "expires_in": 3600}
        tokens = AuthTokens.model_validate(data)
        assert tokens.expires_in == 3600

    def test_paginated_pigeons(self) -> None:
        pigeon_data: dict[str, Any] = {
            "id": "p1",
            "roost_id": "r1",
            "source_ip": "1.2.3.4",
            "request_method": "POST",
            "content_type": "application/json",
            "headers": {},
            "body_json": None,
            "body_raw": None,
            "request_query": None,
            "replayed_from": None,
            "delivery_status": "pending",
            "received_at": "2024-01-01T00:00:00Z",
        }
        data = {"data": [pigeon_data], "next_cursor": "cur_123", "has_more": True}
        page = PaginatedPigeons.model_validate(data)
        assert len(page.data) == 1
        assert page.next_cursor == "cur_123"
        assert page.has_more is True

    def test_delivery_attempt(self) -> None:
        data = {
            "id": "da1",
            "pigeon_id": "p1",
            "destination_id": "d1",
            "status": "failed",
            "attempt_number": 3,
            "response_status": 500,
            "response_body": "Internal Server Error",
            "error_message": "timeout",
            "attempted_at": "2024-01-01T00:01:00Z",
            "next_retry_at": "2024-01-01T00:02:00Z",
        }
        attempt = DeliveryAttempt.model_validate(data)
        assert attempt.status == DeliveryStatus.failed
        assert attempt.attempt_number == 3


class TestRequestModels:
    def test_create_roost_exclude_none(self) -> None:
        req = CreateRoost(name="Test")
        dumped = req.model_dump(exclude_none=True)
        assert dumped == {"name": "Test"}

    def test_update_roost_exclude_unset(self) -> None:
        req = UpdateRoost(name="Updated")
        dumped = req.model_dump(exclude_unset=True)
        assert dumped == {"name": "Updated"}
        assert "description" not in dumped
        assert "is_active" not in dumped

    def test_create_destination_full(self) -> None:
        req = CreateDestination(
            destination_type=DestinationType.url,
            config={"url": "https://example.com"},
            retry_max=3,
        )
        dumped = req.model_dump(exclude_none=True)
        assert dumped["destination_type"] == "url"
        assert dumped["retry_max"] == 3
        assert "filter_expression" not in dumped

    def test_model_validation_error(self) -> None:
        with pytest.raises(ValueError):
            Roost.model_validate({"id": "r1"})  # missing required fields
